<?php

declare(strict_types=1);

namespace apexl\Io\interfaces;

interface messageInterface
{
    public static function addMessage(
        string $name,
        string $type,
        string $content,
        string $icon = '',
        ?int $timeout = null,
        ?string $id = null
    ): void;
}
